//functions for file I/O, parsing, and input error checking

//check that a string will correctly evaluate to a float (returns true even if 
//the number will not be precise)
bool isValidFloat(char* str)
{
  if (str == NULL)
  {
    return 0;
  }
  char c = str[0];
  if(c=='-')
  {
    str++;
    c=*str;
  }
  bool hasdot = false;
  while(c!='\0')
  {
    if(c=='.')
    {
      if(hasdot)
      {
        //false, can't have > 1 decimal points
        return false;
      }
      else
      {
        hasdot = true;
      }
    }
    else
    {
      if(!isdigit(c))
      {
        return false;
      }
    }
    str++;
    c = *str;
  }
  return true;
}

//return true if valid int (only a leading '-' and digits allowed)
//TODO: allow decimals if followed by only 0s
bool isValidInt(char* str)
{
  if (str == (int) NULL)
  {
    return 0;
  }
  char c = str[0];
  if(c=='-')
  {
    str++;
    c=*str;
  }
  while(c!='\0')
  {
    if(isdigit(c))
    {
      str++;
      c = *str;
    }
    else if(c == '.')
    {
      str++;
      c = *str;
      while(c!='\0')
      {
        if(c=='0')
        {
          str++;
          c = *str;
        }
        else
        {
          return false;
        }
      }
    }
    else
    {
      return false;
    }
  }
  return true;
}

//return true if only numeral characters
bool isValidUint(char* str)
{
  if (str == NULL)
  {
    return 0;
  }
  char c = str[0];
  while(c!='\0')
  {
    if(isdigit(c))
    {
      str++;
      c = *str;
    }
    else if(c == '.')
    {
      str++;
      c = *str;
      while(c!='\0')
      {
        if(c=='0')
        {
          str++;
          c = *str;
        }
        else
        {
          return false;
        }
      }
    }
    else
    {
      return false;
    }
  }
  return true;
}

//use file pointers to determine size of textual data, return size or -1 if invalid
long getFileSize(FILE* f)
{
  if(!(fseek(f, 0, SEEK_END)))
  {
    long size = ftell(f);
    fseek(f, 0, SEEK_SET);
    return size;
  }
  else
  {
    return -1;
  }
}

//Copy textual data from file into memory, return a pointer or NULL if invalid
char* wholeFileToMemory(FILE* file)
{
  long size = getFileSize(file) + 1;
  if(size == 0){
    return NULL;
  }

  char* mem = (char*) calloc(size, sizeof(char));
  if(mem==NULL)
  {
    printf("Error allocating memory for file");
    return (char*) NULL;
  }

  fread(mem, sizeof(char), size, file);
  *(mem + size-1) ='\0'; 
  return mem;
}

//Long, annoying function, but it prevents reading whole texture files into memory 
//before parsing them. Expects single-line header
texture loadTexture(char* fileName)
{
  texture tex;
  tex.data = NULL;
  FILE* f = fopen(fileName, "r");
  if(f==NULL)
  {
    std::cerr << "Error: problem opening texture file; check filename." << std::endl; 
    return tex; 
  }
  char buf[100];
  if(fgets(buf, 100, f)==NULL)
  {
    std::cerr << "Error: couldn't read header of texture file." << std::endl; 
    fclose(f);
    return tex;
  }
  char* tokPtr;
  
  char* currentToken = strtok_r(buf, "\n \t", &tokPtr);
  if(!strcmpAny(currentToken, "P3"))
  {
    std::cerr << "Error: It doesn't look like you provided a P3 PPM file as a texture" << std::endl; 
    fclose(f);
    return tex;
  }
  
  currentToken = strtok_r(NULL, "\n \t", &tokPtr);
  if(!isValidUint(currentToken))
  {
    std::cerr << "Error: Texture '" << fileName << "' does not have a valid width." <<std::endl;
    fclose(f);     
    return tex;
  }
  int width = atoi(currentToken);
  
  currentToken = strtok_r(NULL, "\n \t", &tokPtr);
  if(!isValidUint(currentToken))
  {
    std::cerr << "Error: Texture '" << fileName << "' does not have a valid height." <<std::endl;
    fclose(f);     
    return tex;
  }
  int height = atoi(currentToken);
  
  currentToken = strtok_r(NULL, "\n \t", &tokPtr);
  if(!(atoi(currentToken)==255))
  {
    std::cerr << "Error: expected '255' after width and height in a texture file" <<std::endl; 
    fclose(f);    
    return tex;
  }
  int dataSize = height*width*3;
  float* data = (float*) calloc(dataSize, sizeof(float));
  char c;
  int j;
  for (int i=0; i < dataSize; i++)
  {
    c = fgetc(f);
    //advance read
    while( c == ' '|| c == '\n' || c == '\t')
    {
      c = fgetc(f);
    }

    if(c == (int) NULL)
    {
      std::cerr << "Error: early EOF, size header of texture too large for contents" <<std::endl;  
      fclose(f);   
      return tex;
    }
    j = 0;
    //copy (hopefully) numbers into buffer 
    while( c != ' ' && c != '\n' && c != '\t' && c!= EOF)
    {
      buf[j] = c;
      c = fgetc(f);
      j++;
    }
    buf[j] = '\0';
    data[i] = atof(buf);
  }
  fclose(f);
  //malloced so it sticks around 
  tex.width = width;
  tex.height = height;
  tex.data = data;
  return tex;
}

int checkPolyType(char* str)
{
  enum{TEXTURED, SMOOTH_TEXTURED, SMOOTH_MATTE, PLAIN_POLY};
  while(*str != '\n' && *str != '\0' )
  {
    if(*str=='/')
    {
      if(*(str+1)=='/')
      {
        return SMOOTH_MATTE;
      }
      else
      {
        str++;
        while(*str!=' ')
        {
          if(*str == '/')
          {
            return SMOOTH_TEXTURED;
          }
          str++;
        }
		return TEXTURED;
      }
    }
    str++;
  }
  return PLAIN_POLY;
}

//takes a file in memory and a pointer to a renderState, then fills out the state until done
//(and returns 1) or until a problem parsing is encountered (returns 0)
int  parseToRenderState(char* strHead, renderState* state)
{
  bool eyeSet = false, imsizeSet = false, bkgcolorSet = false, 
       updirSet = false, viewdirSet = false, fovvSet = false, mtlcolorSet = false;
  char* tokPtr = NULL;
  char delims[] = " \t\n\r";
  char shadeDelims[] = "/ \t\n\r";
  char* currentToken = strtok_r(strHead, delims, &tokPtr);
  enum{TEXTURED, SMOOTH_TEXTURED, SMOOTH_MATTE, PLAIN_POLY};


  while(currentToken!=NULL)
  {
    
  //parse depending on current keyword; if not a keyword, report error
  //if it is a keyword, check that it is followed by appropriate arguments, report error if not
  //note: we only check for textual information here. For example, this handles alpha characters 
  //mixed in with numbers or negative uints,  but the rendering portion handles nonsense 
  //numerical errors such as an fovv >= 180
    
    if(strcmpAny(currentToken, "eye")) // "eye float float float"
    {
      char* xStr = strtok_r(NULL, delims, &tokPtr);
      char* yStr = strtok_r(NULL, delims, &tokPtr);
      char* zStr = strtok_r(NULL, delims, &tokPtr);

      if(!isValidFloat(xStr))
      {
        std::cerr << "Error: could not parse x value of eye" << std::endl; 
        return 0;
      }
      if(!isValidFloat(yStr))
      {
        std::cerr << "Error: could not parse y value of eye" << std::endl; 
        return 0;
      }
      if(!isValidFloat(zStr))
      {
        std::cerr << "Error: could not parse z value of eye" << std::endl; 
        return 0;
      }
      state->eye.x = atof(xStr);
      state->eye.y = atof(yStr);
      state->eye.z = atof(zStr);
      eyeSet = true;
    }


    else if(strcmpAny(currentToken, "imsize")) // "imsize uint uint" where 0 < uint
    {
      char* wStr = strtok_r(NULL, delims, &tokPtr);
      char* hStr = strtok_r(NULL, delims, &tokPtr);
      
      if(!isValidUint(wStr))
      {
        std::cerr << "Error: could not parse width value of imsize" << std::endl; 
        return 0;
      }
      if(!isValidUint(hStr))
      {
        std::cerr << "Error: could not parse height value of imsize" << std::endl; 
        return 0;
      }
      //add to state, check for correct size on render
      state->imsizeW = atoi(wStr);
      state->imsizeH = atoi(hStr);
      imsizeSet = true;
    }


    else if(strcmpAny(currentToken, "bkgcolor")) // "bkgcolor float float float" where  0<= float <= 1
    {
      char* rStr = strtok_r(NULL, delims, &tokPtr);
      char* gStr = strtok_r(NULL, delims, &tokPtr);
      char* bStr = strtok_r(NULL, delims, &tokPtr);

      if(!isValidFloat(rStr))
      {
        std::cerr << "Error: could not parse r value of bkgcolor" << std::endl; 
        return 0;
      }
      if(!isValidFloat(gStr))
      {
        std::cerr << "Error: could not parse g value of bkgcolor" << std::endl; 
        return 0;
      }
      if(!isValidFloat(bStr))
      {
        std::cerr << "Error: could not parse b value of bkgcolor" << std::endl; 
        return 0;
      }
      state->bkgcolor.r = atof(rStr);
      state->bkgcolor.g = atof(gStr);
      state->bkgcolor.b = atof(bStr);
      //add to state, check for correct formatting upon render
      bkgcolorSet = true;
    }


    else if(strcmpAny(currentToken, "updir"))
    {
      char* xStr = strtok_r(NULL, delims, &tokPtr);
      char* yStr = strtok_r(NULL, delims, &tokPtr);
      char* zStr = strtok_r(NULL, delims, &tokPtr);

      if(!isValidFloat(xStr))
      {
        std::cerr << "Error: could not parse x value of updir" << std::endl; 
        return 0;
      }
      if(!isValidFloat(yStr))
      {
        std::cerr << "Error: could not parse y value of updir" << std::endl; 
        return 0;
      }
      if(!isValidFloat(zStr))
      {
        std::cerr << "Error: could not parse z value of updir" << std::endl; 
        return 0;
      }
      state->updir.x = atof(xStr);
      state->updir.y = atof(yStr);
      state->updir.z = atof(zStr);
      updirSet = true;
    }


    else if(strcmpAny(currentToken, "viewdir"))
    {
      char* xStr = strtok_r(NULL, delims, &tokPtr);
      char* yStr = strtok_r(NULL, delims, &tokPtr);
      char* zStr = strtok_r(NULL, delims, &tokPtr);

      if(!isValidFloat(xStr))
      {
        std::cerr << "Error: could not parse x value of viewdir" << std::endl; 
        return 0;
      }
      if(!isValidFloat(yStr))
      {
        std::cerr << "Error: could not parse y value of viewdir" << std::endl; 
        return 0;
      }
      if(!isValidFloat(zStr))
      {
        std::cerr << "Error: could not parse z value of viewdir" << std::endl; 
        return 0;
      }
      state->viewdir.x = atof(xStr);
      state->viewdir.y = atof(yStr);
      state->viewdir.z = atof(zStr);
      viewdirSet = true;
    }


    else if(strcmpAny(currentToken, "fovv"))
    {
      char* fovvStr = strtok_r(NULL, delims, &tokPtr);
     
      if(!isValidFloat(fovvStr))
      {
        std::cerr << "Error: could not parse value of fovv" << std::endl; 
        return 0;
      }
      //check fovv validity and calculate fovh on render
      state->fovv = atof(fovvStr);
      fovvSet = true;
    }

    //"mtlcolor floatx9 (int) n" where  0<= float <= 1
    else if(strcmpAny(currentToken, "mtlcolor")) 
    {
      char* rStr = strtok_r(NULL, delims, &tokPtr);
      char* gStr = strtok_r(NULL, delims, &tokPtr);
      char* bStr = strtok_r(NULL, delims, &tokPtr);

      char* rSpecStr = strtok_r(NULL, delims, &tokPtr);
      char* gSpecStr = strtok_r(NULL, delims, &tokPtr);
      char* bSpecStr = strtok_r(NULL, delims, &tokPtr);

      char* ambRefStr = strtok_r(NULL, delims, &tokPtr);
      char* difRefStr = strtok_r(NULL, delims, &tokPtr);
      char* specRefStr = strtok_r(NULL, delims, &tokPtr);

      char* nStr = strtok_r(NULL, delims, &tokPtr);

      char* opacityStr = strtok_r(NULL, delims, &tokPtr);
      char* refIndexStr = strtok_r(NULL, delims, &tokPtr);

      if(!isValidFloat(rStr))
      {
        std::cerr << "Error: could not parse r value of mtlcolor" << std::endl; 
        return 0;
      }
      if(!isValidFloat(gStr))
      {
        std::cerr << "Error: could not parse g value of mtlcolor" << std::endl; 
        return 0;
      }
      if(!isValidFloat(bStr))
      {
        std::cerr << "Error: could not parse b value of mtlcolor" << std::endl; 
        return 0;
      }

      if(!isValidFloat(rSpecStr))
      {
        std::cerr << "Error: could not parse red specular highlight value of mtlcolor" << std::endl; 
        return 0;
      }
      if(!isValidFloat(gSpecStr))
      {
        std::cerr << "Error: could not parse green specular highlight value of mtlcolor" << std::endl; 
        return 0;
      }
      if(!isValidFloat(bSpecStr))
      {
        std::cerr << "Error: could not parse blue specular highlight value of mtlcolor" << std::endl; 
        return 0;
      }

      if(!isValidFloat(ambRefStr))
      {
        std::cerr << "Error: could not parse ambient reflectivity value of mtlcolor" << std::endl; 
        return 0;
      }
      if(!isValidFloat(difRefStr))
      {
        std::cerr << "Error: could not parse diffuse reflectivity value of mtlcolor" << std::endl; 
        return 0;
      }
      if(!isValidFloat(specRefStr))
      {
        std::cerr << "Error: could not parse specular reflectivity value of mtlcolor" << std::endl; 
        return 0;
      }

      if(!isValidUint(nStr))
      {
        std::cerr << "Error: could not parse n value of mtlcolor (make sure it's a positive integer)" << std::endl; 
        return 0;
      }

      if(!isValidFloat(opacityStr))
      {
        std::cerr << "Error: could not parse opacity value of mtlcolor" << std::endl; 
        return 0;
      }

      if(!isValidFloat(refIndexStr))
      {
        std::cerr << "Error: could not parse reflectivity index of mtlcolor " << std::endl; 
        return 0;
      }

        float r = atof(rStr);
        float g = atof(gStr);
        float b = atof(bStr);
        float rSpec = atof(rSpecStr);
        float gSpec = atof(gSpecStr);
        float bSpec = atof(bSpecStr);
        float ambRef = atof(ambRefStr);
        float difRef = atof(difRefStr);
        float specRef = atof(specRefStr);
        int n = atoi(nStr);
        float opacity = atof(opacityStr);
        float refIndex = atof(refIndexStr);

        //add to state, check for correct numeric information upon rendering
        state->materials.push_back({r,g,b,rSpec, gSpec, bSpec, ambRef, difRef, specRef, n, 0, opacity, refIndex});
        mtlcolorSet = true;
    }
    //End of keywords, start of parsing geometry

    else if (strcmpAny(currentToken, "light"))
    {
      char* xStr = strtok_r(NULL, delims, &tokPtr);
      char* yStr = strtok_r(NULL, delims, &tokPtr);
      char* zStr = strtok_r(NULL, delims, &tokPtr);
      char* wStr = strtok_r(NULL, delims, &tokPtr);
      char* rStr = strtok_r(NULL, delims, &tokPtr);
      char* gStr = strtok_r(NULL, delims, &tokPtr);
      char* bStr = strtok_r(NULL, delims, &tokPtr);

      if(!isValidFloat(xStr))
      {
        std::cerr << "Error: could not parse x value of a light" << std::endl; 
        return 0;
      }
      if(!isValidFloat(yStr))
      {
        std::cerr << "Error: could not parse y value of a light" << std::endl; 
        return 0;
      }
      if(!isValidFloat(zStr))
      {
        std::cerr << "Error: could not parse z value of a light" << std::endl; 
        return 0;
      }
      if(!isValidUint(wStr))
      {
        std::cerr << "Error: could not parse w value of a light" << std::endl; 
        return 0;
      }
      
      if(!isValidFloat(rStr))
      {
        std::cerr << "Error: could not parse r value of a light" << std::endl; 
        return 0;
      }
      if(!isValidFloat(gStr))
      {
        std::cerr << "Error: could not parse g value of a light" << std::endl; 
        return 0;
      }
      if(!isValidFloat(bStr))
      {
        std::cerr << "Error: could not parse b value of a light" << std::endl; 
        return 0;
      }

      float x = atof(xStr);
      float y = atof(yStr);
      float z = atof(zStr);
      float w = atof(wStr);
      float r = atof(rStr);
      float g = atof(gStr);
      float b = atof(bStr);
      if(w==1)
      {
        state->pointLights.push_back({x,y,z,r,g,b});
      }
      else if(w==0)
      {

        state->directionalLights.push_back({normalizeVec3({x,y,z}),r,g,b});
      }
      else
      {
        std::cerr << "Error: w value of lights should be 0 or 1" << std::endl; 
        return 0;
      }
    }

    else if(strcmpAny(currentToken, "sphere"))
    {
      char* xStr = strtok_r(NULL, delims, &tokPtr);
      char* yStr = strtok_r(NULL, delims, &tokPtr);
      char* zStr = strtok_r(NULL, delims, &tokPtr);
      char* radStr = strtok_r(NULL, delims, &tokPtr);

      if(!mtlcolorSet)
      {
        std::cerr << "Error: You tried to create a sphere before acquiring a material color" << std::endl; 
        return 0; 
      }

      if(!isValidFloat(xStr))
      {
        std::cerr << "Error: could not parse x value of a sphere" << std::endl; 
        return 0;
      }
      if(!isValidFloat(yStr))
      {
        std::cerr << "Error: could not parse y value of a sphere" << std::endl; 
        return 0;
      }
      if(!isValidFloat(zStr))
      {
        std::cerr << "Error: could not parse z value of a sphere" << std::endl; 
        return 0;
      }
      if(!isValidFloat(radStr))
      {
        std::cerr << "Error: could not parse radius value of a sphere" << std::endl; 
        return 0;
      }
      //add to list of spheres:
      float x = atof(xStr);
      float y = atof(yStr);
      float z = atof(zStr);
      float r = atof(radStr);
      int materialIndex = state->materials.size() - 1;
      int textureIndex = state->textures.size() - 1;

      state->spheres.push_back({x,y,z,r,materialIndex, textureIndex});
    }

    else if(strcmpAny(currentToken, "v"))
    {
      char* xStr = strtok_r(NULL, delims, &tokPtr);
      char* yStr = strtok_r(NULL, delims, &tokPtr);
      char* zStr = strtok_r(NULL, delims, &tokPtr);
      if(!isValidFloat(xStr))
      {
        std::cerr << "Error: could not parse x value of a vertex" << std::endl; 
        return 0;
      }
      if(!isValidFloat(yStr))
      {
        std::cerr << "Error: could not parse y value of a vertex" << std::endl; 
        return 0;
      }
      if(!isValidFloat(zStr))
      {
        std::cerr << "Error: could not parse z value of a vertex" << std::endl; 
        return 0;
      }
        float x = atof(xStr);
        float y = atof(yStr);
        float z = atof(zStr);
        state->vertCount++;
        state->verts.push_back({x,y,z});
    }

    else if(strcmpAny(currentToken, "vn"))
    {
      char* v1Str = strtok_r(NULL, delims, &tokPtr);
      char* v2Str = strtok_r(NULL, delims, &tokPtr);
      char* v3Str = strtok_r(NULL, delims, &tokPtr);
      if(!isValidFloat(v1Str))
      {
        std::cerr << "Error: could not parse first value of a vertex normal" << std::endl; 
        return 0;
      }
      if(!isValidFloat(v2Str))
      {
        std::cerr << "Error: could not parse second value of a vertex normal" << std::endl; 
        return 0;
      }
      if(!isValidFloat(v3Str))
      {
        std::cerr << "Error: could not parse third value of a vertex normal" << std::endl; 
        return 0;
      }
      float v1 = atof(v1Str);
      float v2 = atof(v2Str);
      float v3 = atof(v3Str);
      if(v1 == 0 && v2 == 0 && v3 == 0)
      {
        std::cerr << "Error: you provided the 0 vector as a vertex normal" << std::endl; 
        return 0;
      }
      vec3f vn = {v1,v2,v3};
      vn = normalizeVec3(vn);
      state->vertNormsCount++;
      state->vertNorms.push_back(vn);
    }

    else if(strcmpAny(currentToken, "vt"))
    {
      char* xStr = strtok_r(NULL, delims, &tokPtr);
      char* yStr = strtok_r(NULL, delims, &tokPtr);
      if(!isValidFloat(xStr))
      {
        std::cerr << "Error: could not parse the x value of a texture coordinate" << std::endl; 
        return 0;
      }
      if(!isValidFloat(yStr))
      {
        std::cerr << "Error: could not parse the y value of a texture coordinate" << std::endl; 
        return 0;
      }
        float x = atof(xStr);
        float y = atof(yStr);
        state->textureCoordinateCount++;
        state->textureCoordinates.push_back({x, y});
    }

    else if(strcmpAny(currentToken, "f"))
    {
      if(!mtlcolorSet)
      {
        std::cerr << "Error: You tried to create a triangle before acquiring a material color" << std::endl; 
        return 0; 
      }
      
      char *tester = tokPtr;
      int shadeType = checkPolyType(tester);
      int textureCoordinateMax = state->textureCoordinateCount;
      int vertMax = state->vertCount;
	    int vertNormMax = state->vertNormsCount;
      //enum{TEXTURED, SMOOTH_TEXTURED, SMOOTH_MATTE, PLAIN_POLY};

      if(shadeType == TEXTURED)
      {
        char* v1Str = strtok_r(NULL, shadeDelims, &tokPtr);
        char* vt1Str = strtok_r(NULL, shadeDelims, &tokPtr);
        char* v2Str = strtok_r(NULL, shadeDelims, &tokPtr);
        char* vt2Str = strtok_r(NULL, shadeDelims, &tokPtr);
        char* v3Str = strtok_r(NULL, shadeDelims, &tokPtr);
        char* vt3Str = strtok_r(NULL, shadeDelims, &tokPtr);
        
        if(!isValidUint(v1Str)||!isValidUint(v2Str)||!isValidUint(v2Str))
        {
          std::cerr << "Error: irregular vertex number provided" << std::endl; return 0;
        }
        if(!isValidUint(vt1Str)||!isValidUint(vt2Str)||!isValidUint(vt3Str))
        {
          std::cerr << "Error: problem parsing texture coordinates" << std::endl; return 0;
        }
        int vert1 = atoi(v1Str);
        int vert2 = atoi(v2Str);
        int vert3 = atoi(v3Str);
        int tex1 = atoi(vt1Str);
        int tex2 = atoi(vt2Str);
        int tex3 = atoi(vt3Str);
        
        if (vert1 <= 0 || vert1 > vertMax || vert2 <= 0 || vert2 > vertMax || vert3 <= 0 || vert3 > vertMax)
        {
          std::cerr << "Error: tried to create a triangle using a vertex that wasn't in range." << std::endl;
          return 0;
        }
        if (tex1 < 0 || tex1 > textureCoordinateMax || tex2 < 0 || tex2 > textureCoordinateMax || tex3 < 0 || tex3 > textureCoordinateMax )
        {
          std::cerr << "Error: tried to create a triangle using a texture index that wasn't in range." << std::endl;
          return 0;
        }
        polygon p = vertsToPoly(state->verts[vert1], state->verts[vert2], state->verts[vert3]);
        p.materialIndex = state->materials.size() - 1;
        p.textureIndex = state->textures.size() -1;
        p.tex1 = tex1;
        p.tex2 = tex2;
        p.tex3 = tex3;
    		state->polys.push_back(p);
	    } 

      else if (shadeType == SMOOTH_TEXTURED)
      {
        char* v1Str = strtok_r(NULL, shadeDelims, &tokPtr);
        char* vt1Str = strtok_r(NULL, shadeDelims, &tokPtr);
        char* vn1Str = strtok_r(NULL, shadeDelims, &tokPtr);
        char* v2Str = strtok_r(NULL, shadeDelims, &tokPtr);
        char* vt2Str = strtok_r(NULL, shadeDelims, &tokPtr);
        char* vn2Str = strtok_r(NULL, shadeDelims, &tokPtr);
        char* v3Str = strtok_r(NULL, shadeDelims, &tokPtr);
        char* vt3Str = strtok_r(NULL, shadeDelims, &tokPtr);
        char* vn3Str = strtok_r(NULL, shadeDelims, &tokPtr);

        if(!isValidUint(v1Str)||!isValidUint(v2Str)||!isValidUint(v2Str))
        {
          std::cerr << "Error: irregular vertex number provided" << std::endl; return 0;
        }
        if(!isValidUint(vn1Str)||!isValidUint(vn2Str)||!isValidUint(vn3Str))
        {
          std::cerr << "Error: irregular vertex number provided" << std::endl; return 0;
        }
        if(!isValidUint(vt1Str)||!isValidUint(vt2Str)||!isValidUint(vt3Str))
        {
          std::cerr << "Error: problem parsing texture coordinates" << std::endl; return 0;
        }
        int vert1 = atoi(v1Str);
        int vert2 = atoi(v2Str);
        int vert3 = atoi(v3Str);
        int norm1 = atoi(vn1Str);
        int norm2 = atoi(vn2Str);
        int norm3 = atoi(vn3Str);
        int tex1 = atoi(vt1Str);
        int tex2 = atoi(vt2Str);
        int tex3 = atoi(vt3Str);

        if (vert1 <= 0 || vert1 > vertMax || vert2 <= 0 || vert2 > vertMax || vert3 <= 0 || vert3 > vertMax)
        {
          std::cerr << "Error: tried to create a triangle using a vertex that wasn't in range." << std::endl;
          return 0;
        }
        if (norm1 < 0 || norm1 > vertNormMax || norm2 < 0 || norm2 > vertNormMax || norm3 < 0 || norm3 > vertNormMax)
        {
          std::cerr << "Error: tried to create a triangle using a vertex normal that wasn't in range." << std::endl;
          return 0;
        }
        if (tex1 < 0 || tex1 > textureCoordinateMax || tex2 < 0 || tex2 > textureCoordinateMax || tex3 < 0 || tex3 > textureCoordinateMax )
        {
          std::cerr << "Error: tried to create a triangle using a texture coordinate that wasn't in range." << std::endl;
          return 0;
        }
        polygon p = vertsToPoly(state->verts[vert1], state->verts[vert2], state->verts[vert3]);
        p.materialIndex = state->materials.size() - 1;
        p.textureIndex = state->textures.size() - 1;
        p.vert0Norm =  norm1;
        p.vert1Norm =  norm2;
        p.vert2Norm =  norm3;
        p.tex1 = tex1;
        p.tex2 = tex2;
        p.tex3 = tex3;
        state->polys.push_back(p);
      }  
     
      else if(shadeType == SMOOTH_MATTE)
      {
        char* v1Str = strtok_r(NULL, shadeDelims, &tokPtr);
        char* vn1Str = strtok_r(NULL, shadeDelims, &tokPtr);
        char* v2Str = strtok_r(NULL, shadeDelims, &tokPtr);
        char* vn2Str = strtok_r(NULL, shadeDelims, &tokPtr);
        char* v3Str = strtok_r(NULL, shadeDelims, &tokPtr);
        char* vn3Str = strtok_r(NULL, shadeDelims, &tokPtr);

        if(!isValidUint(v1Str)||!isValidUint(v2Str)||!isValidUint(v2Str))
        {
          std::cerr << "Error: irregular vertex number provided" << std::endl; return 0;
        }
        if(!isValidUint(vn1Str)||!isValidUint(vn2Str)||!isValidUint(vn3Str))
        {
          std::cerr << "Error: irregular vertex number provided" << std::endl; return 0;
        }
        int vert1 = atoi(v1Str);
        int vert2 = atoi(v2Str);
        int vert3 = atoi(v3Str);
        int norm1 = atoi(vn1Str);
        int norm2 = atoi(vn2Str);
        int norm3 = atoi(vn3Str);
        if (vert1 <= 0 || vert1 > vertMax || vert2 <= 0 || vert2 > vertMax || vert3 <= 0 || vert3 > vertMax)
        {
          std::cerr << "Error: tried to create a triangle using a vertex that wasn't in range." << std::endl;
          return 0;
        }
        if (norm1 < 0 || norm1 > vertNormMax || norm2 < 0 || norm2 > vertNormMax || norm3 < 0 || norm3 > vertNormMax)
        {
          std::cerr << "Error: tried to create a triangle using a vertex normal that wasn't in range." << std::endl;
          return 0;
        }
        polygon p = vertsToPoly(state->verts[vert1], state->verts[vert2], state->verts[vert3]);
        p.materialIndex = state->materials.size() - 1;
        // textureIndex already initialized to 0
        p.vert0Norm =  norm1;
        p.vert1Norm =  norm2;
        p.vert2Norm =  norm3;
        state->polys.push_back(p);
      }

      else //expect a PLAIN_POLY (no texture, no smooth shading)
      {
        char* v1Str = strtok_r(NULL, shadeDelims, &tokPtr);
        char* v2Str = strtok_r(NULL, shadeDelims, &tokPtr);
        char* v3Str = strtok_r(NULL, shadeDelims, &tokPtr);
      
        if(!isValidUint(v1Str)||!isValidUint(v2Str)||!isValidUint(v3Str))
        {
          std::cerr << "Error: irregular vertex number provided" << std::endl; return 0;
        }
        int vert1 = atoi(v1Str);
        int vert2 = atoi(v2Str);
        int vert3 = atoi(v3Str);
        if (vert1 <= 0 || vert1 > vertMax || vert2 <= 0 || vert2 > vertMax || vert3 <= 0 || vert3 > vertMax)
        {
          std::cerr << "Error: tried to create a triangle using a vertex that wasn't in range." << std::endl;
          return 0;
        }
        polygon p = vertsToPoly(state->verts[vert1], state->verts[vert2], state->verts[vert3]);
        p.materialIndex = state->materials.size() - 1;
        // textureIndex already initialized to 0
        state->polys.push_back(p);
      }   
    }

    else if(strcmpAny(currentToken, "texture"))
    {
      char* filename = strtok_r(NULL, delims, &tokPtr);
      texture tex = loadTexture(filename);
      if(!tex.data)
      {
        std::cerr << "Error: problem loading texture, see above" << std::endl; 
       return 0;
      }
      state->textures.push_back(tex);
    }
    else if(strcmpAny(currentToken, "parallel"))
    {
      state->parallelMode = 1;
    }


    /*
    else if(strcmpAny(currentToken, "ellipsoid"))
    {
      char* xStr = strtok_r(NULL, delims, &tokPtr);
      char* yStr = strtok_r(NULL, delims, &tokPtr);
      char* zStr = strtok_r(NULL, delims, &tokPtr);
      char* xRadStr = strtok_r(NULL, delims, &tokPtr);
      char* yRadStr = strtok_r(NULL, delims, &tokPtr);
      char* zRadStr = strtok_r(NULL, delims, &tokPtr);

      if(!mtlcolorSet)
      {
        std::cerr << "Error: You tried to create a sphere before acquiring a material color" << std::endl; 
        return 0; 
      }

      if(!isValidFloat(xStr))
      {
        std::cerr << "Error: could not parse x value of an ellipse" << std::endl; 
        return 0;
      }
      if(!isValidFloat(yStr))
      {
        std::cerr << "Error: could not parse y value of an ellipse" << std::endl; 
        return 0;
      }
      if(!isValidFloat(zStr))
      {
        std::cerr << "Error: could not parse z value of an ellipse" << std::endl; 
        return 0;
      }
      if(!isValidFloat(xRadStr))
      {
        std::cerr << "Error: could not parse x radius value of an ellipse" << std::endl; 
        return 0;
      }
      if(!isValidFloat(yRadStr))
      {
        std::cerr << "Error: could not parse y radius value of an ellipse" << std::endl; 
        return 0;
      }
      if(!isValidFloat(zRadStr))
      {
        std::cerr << "Error: could not parse z radius value of an ellipse" << std::endl; 
        return 0;
      }
      //add to list of ellipses:
      float x = atof(xStr);
      float y = atof(yStr);
      float z = atof(zStr);
      float xRad = atof(xRadStr);
      float yRad = atof(yRadStr);
      float zRad = atof(zRadStr);
      int materialIndex = state->materials.size() - 1;
      state->ellipses.push_back({x,y,z,xRad, yRad, zRad, materialIndex});
    }
    */
    
    else
    {
      std::cerr << "Error: you passed an unrecognized keyword" << std::endl; 
      return 0;
    }

    //setup string for next loop
    currentToken = strtok_r(NULL, delims, &tokPtr);
  }

  //done reading file, check if whole state was initialized
  bool wholeStateSet = true;
  if(!eyeSet)
  {
    std::cerr << "Error: please initialize eye in your input file." << std::endl; 
    wholeStateSet = false;
  }
  if(!imsizeSet)
  {
    std::cerr << "Error: please initialize imsize in your input file." << std::endl;
    wholeStateSet = false; 
  }
  if(!bkgcolorSet)
  {
    std::cerr << "Error: please initialize bkgcolor in your input file." << std::endl;
    wholeStateSet = false; 
  }
  if(!updirSet)
  {
    std::cerr << "Error: please initialize updir in your input file." << std::endl;
    wholeStateSet = false; 
  }
  if(!viewdirSet)
  {
    std::cerr << "Error: please initialize viewdir in your input file." << std::endl;
    wholeStateSet = false; 
  }
  if(!fovvSet)
  {
    std::cerr << "Error: please initialize fovv in your input file." << std::endl; 
    wholeStateSet = false;
  }
  if(!mtlcolorSet)
  {
    std::cerr << "Error: please initialize at least one mtlcolor in your input file." << std::endl;
    wholeStateSet = false; 
  }
  if(wholeStateSet) { return 1; } //It worked!
  else {return 0;}  //error encountered
}



