#include"renderer.h"

int checkStateVariables(renderState* state)
{
  //Even 8192 x 8192 is a very large file, more than that is a pretty huge file for this format
  if(state->imsizeW > MAX_PPM_DIM || state->imsizeH > MAX_PPM_DIM)
  {
    std::cerr << "Error: Please provide output image dimensions such that each is less than " << MAX_PPM_DIM << " pixels." << std::endl; 
    return 0;
  }

  if(state->imsizeW < 1 || state->imsizeH < 1)
  {
    std::cerr << "Error: please supply greater values for the output image size that are greater than 0." << std::endl; 
    return 0;
  }

  for (unsigned int i = 0; i < state->materials.size(); i++)
  {
    if(state->materials[i].r < 0 || state->materials[i].r > 1 ||
       state->materials[i].g < 0 || state->materials[i].g > 1 ||
       state->materials[i].b < 0 || state->materials[i].b > 1)
    {
     std::cerr << "Error: please only supply material colors between 0 and 1 (inclusive)." << std::endl;  
     return 0;
    }

    if(state->materials[i].rSpec < 0 || state->materials[i].rSpec > 1 ||
       state->materials[i].gSpec < 0 || state->materials[i].gSpec > 1 ||
       state->materials[i].bSpec < 0 || state->materials[i].bSpec > 1)
    {
     std::cerr << "Error: please only supply specular highlight colors between 0 and 1 (inclusive)." << std::endl;  
     return 0;
    }

    if(state->materials[i].ambRef < 0  || state->materials[i].ambRef > 1 ||
       state->materials[i].difRef < 0  || state->materials[i].difRef > 1 ||
       state->materials[i].specRef < 0 || state->materials[i].specRef > 1)
    {
     std::cerr << "Error: please only supply material reflective values between 0 and 1 (inclusive)." << std::endl;  
     return 0;
    }
    if(state->materials[i].n<0)
    {
      std::cerr << "Error: please only supply material n value >= 0." << std::endl;  
      return 0;
    }
    return 1;
  }

  if(state->bkgcolor.r < 0 || state->bkgcolor.r > 1 ||
     state->bkgcolor.g < 0 || state->bkgcolor.g > 1 ||
     state->bkgcolor.b < 0 || state->bkgcolor.b > 1)
  {
    std::cerr << "Error: please only supply bkgcolor colors between 0 and 1 (inclusive)." << std::endl;  
    return 0;
  }
  if(state->fovv < 0 || state->fovv >= 180)
  {
    std::cerr << "Error: please supply a fovv greater than 0 and less than 180" << std::endl; 
    return 0;
  }

  for(unsigned int i = 0; i < state->directionalLights.size(); i++)
  {
    if(state->directionalLights[i].c.r < 0 || state->directionalLights[i].c.r > 1 ||
       state->directionalLights[i].c.g < 0 || state->directionalLights[i].c.g > 1 ||
       state->directionalLights[i].c.b < 0 || state->directionalLights[i].c.b > 1 )
    {
      std::cerr << "Error: please only pass use colors between 0 and 1 (inclusive) for directional lights" << std::endl; 
      return 0;
    }
  }

  for(int i = 0; i < state->pointLights.size(); i++)
  {
    if(state->pointLights[i].c.r < 0 || state->pointLights[i].c.r > 1 ||
	   state->pointLights[i].c.g < 0 || state->pointLights[i].c.g > 1 ||
	   state->pointLights[i].c.b < 0 || state->pointLights[i].c.b > 1)
    {
      std::cerr << "Error: please only pass use colors between 0 and 1 (inclusive) for directional lights" << std::endl; 
      return 0;
    }
  }

  for(int i = 0; i < state->spheres.size(); i++)
  {
    if(state->spheres[i].radius <= 0)
    {
      std::cerr << "Error: please only create spheres with a radius greater than 0." << std::endl; 
      return 0;
    }
  }

  for(int i = 0; i < state->ellipses.size(); i++)
  {
    if(state->ellipses[i].xRad <= 0 ||
       state->ellipses[i].yRad <= 0||
       state->ellipses[i].zRad <= 0)
    {
      std::cerr << "Error: please only create ellipses such that each radii is greater than 0." << std::endl; 
      return 0;
    }
  }
}

//Sends multiple shadow rays toward scattered points around a light source, simulating soft shadows
//Called instead of pointShadowRay when SOFT_SHADOWS flag is turned on in main file
float pointShadowBundle(renderState* state, point3f origin, point3f lightCenter)
{
  float shadow=0;
  for(int i = 0; i < RAYS_PER_SHADOW_BUNDLE; i++)
  {
    //Cast a ray toward the light point with a random radius added to it
    float x = RandomFloat(-POINT_LIGHT_RADIUS,POINT_LIGHT_RADIUS) + lightCenter.x;
    float y = RandomFloat(-POINT_LIGHT_RADIUS,POINT_LIGHT_RADIUS) + lightCenter.y;
    float z = RandomFloat(-POINT_LIGHT_RADIUS,POINT_LIGHT_RADIUS) + lightCenter.z;
    
    point3f lightPoint = {x,y,z};
    vec3f lightDir = normalizeVec3({x - origin.x, y - origin.y, z - origin.z});

    int notInShadow=1;
    float lightDistance = sqrt(innerProduct(lightPoint - origin, lightPoint - origin));
    for(unsigned int j = 0; j < state->spheres.size(); j++)
    {
      //Standard sphere intersection test, but either add 1 or 0 to running intersection count
      notInShadow = 1;
      vec3f sphereCenterDir = (origin - state->spheres[j].center);
      float B = 2*innerProduct(lightDir, sphereCenterDir);
      //C = (sphereCenterDir dot sphereCenterDir) - R*R
      float C = innerProduct(sphereCenterDir, sphereCenterDir) - (state->spheres[j].radius * state->spheres[j].radius); 
      float discriminant = (B*B)-(4*C);
      //otherwise, should hit; use distance to possibly occlude other shapes
      if(discriminant > 0)
      {
        //two solutions 
        float solutionOne = (-B + sqrt(discriminant))/2;
        float solutionTwo = (-B - sqrt(discriminant))/2;
        
        if(solutionOne < lightDistance && solutionOne > SHADOW_EPSILON )
        {
          notInShadow = 0;
          break;
        }
        else if (solutionTwo < lightDistance && solutionTwo > SHADOW_EPSILON)
        {
          notInShadow = 0;
          break;
        }
      }
    }
  if(notInShadow)
  {    
    for(int k = 0; k < state->polys.size(); k++)
    {
      polygon p = state->polys[k];
      float denom = innerProduct(p.n, lightDir);
      //if ray is not parallel
      if (denom > INTERSECT_EPSILON || denom < -INTERSECT_EPSILON)
      {
        float t = -((p.n.x*origin.x +  p.n.y*origin.y + p.n.z*origin.z) + p.d)/denom;
        point3f intersectPoint = origin + t*lightDir;
        if(t > SHADOW_EPSILON && t < lightDistance)
        {
          vec3f e3 =  intersectPoint - p.p1;
          vec3f e4 = intersectPoint - p.p2;
          vec3f temp = crossProduct(e3, e4);
          float alpha = .5*sqrt(innerProduct(temp,temp))/p.area;
          temp = crossProduct(e4, p.e2);
          float beta = .5*sqrt(innerProduct(temp, temp))/p.area;
          temp = crossProduct(p.e1, e3);
          float gamma = .5*sqrt(innerProduct(temp,temp))/p.area;
          if(alpha + beta + gamma - 1 < INTERSECT_EPSILON)
          {
            notInShadow = 0;
            break;
          }
        }
      }
    }
  }
  shadow = shadow + notInShadow;
  }
  //average the results and return that 0 to 1 float
  return shadow/(float)RAYS_PER_SHADOW_BUNDLE; 
}

//Used to detect intersections for hard shadows
//Return 1 if no sphere collision is detected, 0 if one was detected
float pointShadowray(renderState* state, point3f intersectPoint, vec3f lightDir, point3f lightPoint)
{ 
  float shadow = 1;
  float lightDistance = sqrt(innerProduct(lightPoint - intersectPoint, lightPoint - intersectPoint));
  for(int k = 0; k < state->spheres.size(); k++)
  {
    sphere s = state->spheres[k];
    //A == 1, since ray is normalized and A = innerProduct(lightDir, lightDir) 
    //B = 2 * lightDir dot (intersectPoint - sphereCenter)
    vec3f sphereCenterDir = (intersectPoint - s.center);
    float B = 2*innerProduct(lightDir, sphereCenterDir);
    //C = (sphereCenterDir dot sphereCenterDir) - R*R
    float C = innerProduct(sphereCenterDir, sphereCenterDir) - (s.radius * s.radius); 
    float discriminant = (B*B)-(4*C);
    //otherwise, should hit; use distance to possibly occlude other shapes
    if(discriminant > 0)
    {
      //two solutions 
      float solutionOne = (-B + sqrt(discriminant))/2;
      float solutionTwo = (-B - sqrt(discriminant))/2;
      if(solutionOne > SHADOW_EPSILON && solutionOne < lightDistance)
      {
		    shadow -= state->materials[ state->spheres[k].materialIndex].opacity;
        if(shadow<=0)
        {
          return 0;
        }
      }
      else if(solutionTwo > SHADOW_EPSILON && solutionTwo < lightDistance)
      {
		    shadow -= state->materials[state->spheres[k].materialIndex].opacity;
        if(shadow<=0)
        {
          return 0;
        }
      }
    }
  }

  for(int k = 0; k < state->polys.size(); k++)
  {
    polygon p = state->polys[k];
    float denom = innerProduct(p.n, lightDir);
    //if ray is not parallel
    if (denom > INTERSECT_EPSILON || denom < -INTERSECT_EPSILON)
    {
      float t = -((p.n.x*intersectPoint.x +  p.n.y*intersectPoint.y + p.n.z*intersectPoint.z) + p.d)/denom;
      point3f planeCollision = intersectPoint + t*lightDir;
      if(t > SHADOW_EPSILON )
      {
        if(t<lightDistance)
        {
          vec3f e3 =  planeCollision - p.p1;
          vec3f e4 = planeCollision - p.p2;
          vec3f temp = crossProduct(e3, e4);
          float alpha = .5*sqrt(innerProduct(temp,temp))/p.area;
          temp = crossProduct(e4, p.e2);
          float beta = .5*sqrt(innerProduct(temp, temp))/p.area;
          temp = crossProduct(p.e1, e3);
          float gamma = .5*sqrt(innerProduct(temp,temp))/p.area;
          if(alpha + beta + gamma - 1 < INTERSECT_EPSILON)
          {
            shadow -= state->materials[state->polys[k].materialIndex].opacity;
            if(shadow<=0)
            {
              return 0;
            }
          }
        }
      }
    }
  }
  return shadow;//no intersection detected
}

//for directional lights, any positive intersection means the pixel is shadowed, so return 0
//if an intersection is found further away than some positive epsilon (no need to check distance)
float directionalShadowray(renderState* state, point3f intersectPoint, vec3f lightDir)
{
  float shadow = 1;
  for(unsigned int k = 0; k < state->spheres.size(); k++)
  {
    sphere s = state->spheres[k];
    //A == 1, since ray is normalized and A = innerProduct(lightDir, lightDir) 
    //B = 2 * lightDir dot (intersectPoint - sphereCenter)
    vec3f sphereCenterDir = (intersectPoint - s.center);
    float B = 2*innerProduct(lightDir, sphereCenterDir);
    //C = (sphereCenterDir dot sphereCenterDir) - R*R
    float C = innerProduct(sphereCenterDir, sphereCenterDir) - (s.radius * s.radius); 
    float discriminant = (B*B)-(4*C);

    //otherwise, should hit; use distance to possibly occlude other shapes
    if(discriminant > 0)
    {
      //two solutions 
      float solutionOne = (-B + sqrt(discriminant))/2;
      float solutionTwo = (-B - sqrt(discriminant))/2;
      if(solutionOne > SHADOW_EPSILON || solutionTwo > SHADOW_EPSILON)
      {
        shadow -= state->materials[state->spheres[k].materialIndex].opacity;
        if(shadow<=0)
        {
          return 0;
        }
      }
    }
  }
  //POLYGON TESTING
  for(int k = 0; k < state->polys.size(); k++)
  {
    polygon p = state->polys[k];
    float denom = innerProduct(p.n, lightDir);
    //if ray is not parallel
    if (denom > INTERSECT_EPSILON || denom < -INTERSECT_EPSILON)
    {
      float t = -((p.n.x*intersectPoint.x +  p.n.y*intersectPoint.y + p.n.z*intersectPoint.z) + p.d)/denom;
      intersectPoint = intersectPoint + t*lightDir;
      if(t > SHADOW_EPSILON )
      {
        vec3f e3 =  intersectPoint - p.p1;
        vec3f e4 = intersectPoint - p.p2;
        vec3f temp = crossProduct(e3, e4);
        float alpha = .5*sqrt(innerProduct(temp,temp))/p.area;
        temp = crossProduct(e4, p.e2);
        float beta = .5*sqrt(innerProduct(temp, temp))/p.area;
        temp = crossProduct(p.e1, e3);
        float gamma = .5*sqrt(innerProduct(temp,temp))/p.area;
        if(alpha + beta + gamma - 1 < INTERSECT_EPSILON)
        {
          shadow -= state->materials[state->polys[k].materialIndex].opacity;
          if(shadow<=0)
          {
            return 0;
          }
        }
      }
    }
  }
  return shadow;
}

color phongLighting(vec3f viewRay, point3f intersectPoint, vec3f surfaceNorm, material mat, renderState* state, int traceNumber, float refractionIndexes[MAX_PHONG_RECURSIONS+1])
{
  //Start with ambient terms for current material
  float red = mat.ambRef*mat.r;
  float green = mat.ambRef*mat.g;
  float blue = mat.ambRef*mat.b;
  float shadow; //value returned from shadow rays
  int insideObj = false;
  color recurColorResult = {0,0,0};
  color transparentRecurColorResult = {0,0,0};
  //reflections:
  
  vec3f incidentRay = -(viewRay);
  float incidenceCos = innerProduct(surfaceNorm, incidentRay);
  if (incidenceCos < 0 )
  { // must be inside a surface, set transmitted refraction index to air
    refractionIndexes[traceNumber] = 1;
    insideObj = true;
    surfaceNorm = -surfaceNorm;
    incidenceCos = innerProduct(surfaceNorm, incidentRay);
  }
  else
  { // not inside a surface, entering mat's refraction index
    refractionIndexes[traceNumber] = mat.refIndex;
  }

  float f0 = (refractionIndexes[traceNumber] - refractionIndexes[traceNumber-1]) / (refractionIndexes[traceNumber] + refractionIndexes[traceNumber-1]);
  f0 = f0*f0;
  float fresnelRef = (f0) + (1-f0)* pow((1-incidenceCos),5);
  
  vec3f refRay = ((2*incidenceCos)*(surfaceNorm)) - incidentRay;
  
  if(traceNumber <= MAX_PHONG_RECURSIONS)
  {
	  recurColorResult = closestShapeAndLighting(state, intersectPoint, refRay, traceNumber, refractionIndexes);
    recurColorResult = fresnelRef * recurColorResult;
  }
  

  if(mat.opacity < 1 && traceNumber <= MAX_PHONG_RECURSIONS)
  {
    //(ni/nt)^2
    float transparentRayCos = pow((refractionIndexes[traceNumber-1]/refractionIndexes[traceNumber]),2);
    //(1-(cos theta i)^2)
    transparentRayCos *= (1 - (incidenceCos*incidenceCos));
    transparentRayCos = 1 - transparentRayCos;
    if(transparentRayCos >= 0)
    {
      transparentRayCos = sqrt(transparentRayCos);
      vec3f transparentRayA = -surfaceNorm;
      transparentRayA = transparentRayCos*transparentRayA;
      vec3f transparentRayB = (refractionIndexes[traceNumber-1]/refractionIndexes[traceNumber])*((incidenceCos*surfaceNorm)-incidentRay);
	  transparentRecurColorResult = transparentRecurColorResult + (1 - fresnelRef)*(1 - mat.opacity)*closestShapeAndLighting(state, intersectPoint, transparentRayA + transparentRayB, traceNumber, refractionIndexes);
    }
  }


  //for each light, add the light intensity * (diffuse + specular)
  for (int k = 0; k < state->pointLights.size(); k++)
  {
    vec3f lightDir = state->pointLights[k].center - intersectPoint;
    lightDir = normalizeVec3(lightDir);
    float diffuseCos = innerProduct(surfaceNorm, lightDir);
    if(diffuseCos < 0)
    { 
      diffuseCos = 0;
    }
    //light intensities
    float lightRed = state->pointLights[k].c.r;
    float lightGreen = state->pointLights[k].c.g;
    float lightBlue = state->pointLights[k].c.b;
    vec3f halfway = -(viewRay) + lightDir;
    halfway = normalizeVec3(halfway);
    float specTerm = innerProduct(surfaceNorm, halfway);
    if(specTerm>0)
    {
      specTerm = pow(specTerm,mat.n) * mat.specRef;
    }
    else
    {
      specTerm = 0;
    }
    #if SOFT_SHADOWS
    shadow = pointShadowBundle(state, intersectPoint, state->pointLights[k].center);
    #else
    shadow = pointShadowray(state, intersectPoint, lightDir, state->pointLights[k].center);
    #endif
    red = red + shadow*lightRed*(mat.difRef*mat.r*diffuseCos + specTerm*mat.rSpec);
    green = green + shadow*lightGreen*(mat.difRef*mat.g*diffuseCos + specTerm*mat.gSpec);
    blue = blue + shadow*lightBlue*(mat.difRef*mat.b*diffuseCos + specTerm*mat.bSpec);
  }

  for (int k = 0; k < state->directionalLights.size(); k++)
  {
    vec3f lightDir = -state->directionalLights[k].direction;
    float diffuseCos = innerProduct(surfaceNorm, lightDir);
    if(diffuseCos < 0) 
    {
      diffuseCos = 0;
    }
    //light intensities
    float lightRed = state->directionalLights[k].c.r;
    float lightGreen = state->directionalLights[k].c.g;
    float lightBlue = state->directionalLights[k].c.b;
    vec3f halfway = -(viewRay) + lightDir;
    halfway = normalizeVec3(halfway);
    float specTerm = innerProduct(surfaceNorm, halfway);
    if(specTerm>0)
    {
      specTerm = pow(specTerm,mat.n) * mat.specRef;
    }
    else
    {
      specTerm = 0;
    }
    //Values are unchanged if in directional shadow
    shadow = directionalShadowray(state, intersectPoint, lightDir);
    red = red + shadow*lightRed*(mat.difRef*mat.r*diffuseCos + specTerm*mat.rSpec);
    green = green + shadow*lightGreen*(mat.difRef*mat.g*diffuseCos + specTerm*mat.gSpec);
    blue = blue + shadow*lightBlue*(mat.difRef*mat.b*diffuseCos + specTerm*mat.bSpec);
  }
  //end of lighting/shadow calculations, clamp to 1
  red = red + recurColorResult.r + transparentRecurColorResult.r;
  green = green + recurColorResult.g + transparentRecurColorResult.g;
  blue = blue + recurColorResult.b + transparentRecurColorResult.g;
  if (red > 1) red = 1;
  if (green>1) green=1;
  if (blue>1) blue=1;
  return {red, green, blue};
}

color closestShapeAndLighting(renderState* state, point3f origin, vec3f viewRay, int traceNumber, float refractionIndexes[MAX_PHONG_RECURSIONS+1])
{
  traceNumber++;

  if(traceNumber>=MAX_PHONG_RECURSIONS)
  {
    return {0,0,0};
  }
  sphere closestSphere;
  polygon closestPolygon;
  float closestAlpha, closestBeta, closestGamma;
  float closestT = FLT_MAX; 
  int materialIndex;
  color resultColor = state->bkgcolor;
  int closestShapeType = NONE;
  //via implicit equation: A = 1 (for spheres); B = 2*viewRay dot (origin - sphere.center); 
  // C = (origin - sphere.center) dot (origin - sphere.center) - r*r and discriminant = B*B - 4AC
  //consider the closest intersection to be the max float value away, with a color matching the background
  //Use T to denote the magnitude term in: origin + viewRay*T
  //SPHERE TESTING
  for(int k = 0; k < state->spheres.size(); k++)
  {
    sphere s = state->spheres[k];
    //A == 1, since ray is normalized and A = innerProduct(viewRay, viewRay) 
    //B = 2*viewRay dot (origin - sphereCenter)
    vec3f originToCenter = (origin - s.center);
    float B = 2*innerProduct(viewRay, originToCenter);
    //C = ((origin - sphereCenter) dot (origin - sphereCenter)) - R*R
    float C = innerProduct(originToCenter, originToCenter) - (s.radius * s.radius); 
    float discriminant = (B*B)-(4*C);
    if(discriminant < 0)
    {
      //do nothing, ray misses
    }
    //otherwise, should hit; use distance to possibly occlude other shapes
    else if(discriminant > 0)
    {
      //two solutions: 
      float solutionOne = (-B + sqrt(discriminant)) / 2.0f;
      float solutionTwo = (-B - sqrt(discriminant)) / 2.0f;
      if(solutionOne > INTERSECT_EPSILON && solutionOne < closestT)
      {
        closestT = solutionOne;
        closestShapeType = SPHERE;
        closestSphere = s;
      }
      if(solutionTwo > INTERSECT_EPSILON && solutionTwo < closestT)
      {
        closestT = solutionTwo;
        closestShapeType = SPHERE;
        closestSphere = s;
      }
    }
    else if(discriminant == 0)
    {
      float solution = (-B)/2;
      if( solution > INTERSECT_EPSILON && solution < closestT )
      {
        closestT = solution;
        closestShapeType = SPHERE;
        closestSphere = s;
      }
    }
    else
    {
      std::cerr << "Error: sphere detection problem (this should never happen)" << std::endl; 
    }
  }//end of sphere testing

  //POLYGON TESTING
  for(int k = 0; k < state->polys.size(); k++)
  {
    polygon p = state->polys[k];
    float denom = innerProduct(p.n, viewRay);
    //if ray is not parallel
    if (denom > INTERSECT_EPSILON || denom < -INTERSECT_EPSILON)
    {
      float t = -((p.n.x*origin.x +  p.n.y*origin.y + p.n.z*origin.z) + p.d)/denom;
      point3f intersectPoint = origin + t*viewRay;
      //TODO: checking t < closestT is okay here, right?
      if(t > INTERSECT_EPSILON && t < closestT)
      {
        vec3f e3 =  intersectPoint - p.p1;
        vec3f e4 = intersectPoint - p.p2;
        vec3f temp = crossProduct(e3, e4);
        float alpha = .5*sqrt(innerProduct(temp,temp))/p.area;
        temp = crossProduct(e4, p.e2);
        float beta = .5*sqrt(innerProduct(temp, temp))/p.area;
        temp = crossProduct(p.e1, e3);
        float gamma = .5*sqrt(innerProduct(temp,temp))/p.area;
        if(alpha + beta + gamma - 1 < INTERSECT_EPSILON)
        {
          if(p.vert0Norm>=0 && p.vert1Norm>=0 && p.vert2Norm>=0)
          {
              vec3f shadingNorm = alpha*state->vertNorms[p.vert0Norm] + 
              beta*state->vertNorms[p.vert1Norm]+ gamma*state->vertNorms[p.vert2Norm];
              p.shadingNorm = normalizeVec3(shadingNorm);
          }
          closestAlpha = alpha;
          closestBeta = beta;
          closestGamma = gamma;
          closestT = t;
          closestShapeType = POLYGON;
          closestPolygon = p;
        }
      }
    }
  } //end of poly test
  material mat;
  point3f intersectPoint;
  vec3f surfaceNorm;
  float u, v;
  //check which type of shape was closest 
  switch(closestShapeType)
  {
    case NONE: 
      if(traceNumber==1){resultColor = state->bkgcolor; }
      else{resultColor = {0,0,0};}
      break;
    case POLYGON: 
      mat = state->materials[closestPolygon.materialIndex];
      intersectPoint = origin + (closestT*viewRay);
      if(closestPolygon.textureIndex)
      {
        texture tex = state->textures[closestPolygon.textureIndex];
        u = closestAlpha * state->textureCoordinates[closestPolygon.tex1].x + 
            closestBeta * state->textureCoordinates[closestPolygon.tex2].x +
            closestGamma * state->textureCoordinates[closestPolygon.tex3].x;
        v = closestAlpha * state->textureCoordinates[closestPolygon.tex1].y +
          closestBeta * state->textureCoordinates[closestPolygon.tex2].y +
          closestGamma * state->textureCoordinates[closestPolygon.tex3].y;
        //compute rgb from texture and set material to that 
         int i = round(u*(tex.width -1));
         int j = round(v*(tex.height -1));
         i = j*(tex.width*3) + i*3;
         //i=i*3;
         mat.r = (float)tex.data[i]/255.0;
         mat.g = (float)tex.data[i+1]/255.0;
         mat.b = (float)tex.data[i+2]/255.0;
      }
      //If vert normals are just defined as the face, not provided
      if(closestPolygon.vert0Norm < 0 || closestPolygon.vert1Norm < 0 || closestPolygon.vert2Norm < 0)
      {
    
        resultColor = phongLighting(viewRay, intersectPoint, closestPolygon.shadingNorm, mat, state, traceNumber, refractionIndexes); 
      }
      else
      {
        vec3f shadingNorm = 
          closestAlpha*state->vertNorms[closestPolygon.vert0Norm] + 
          closestBeta*state->vertNorms[closestPolygon.vert1Norm] + 
          closestGamma*state->vertNorms[closestPolygon.vert2Norm];
        shadingNorm = normalizeVec3(shadingNorm);
        resultColor = phongLighting(viewRay, intersectPoint, shadingNorm, mat, state, traceNumber, refractionIndexes);
      }
      break;
    case SPHERE: 
      mat = state->materials[closestSphere.materialIndex];
      intersectPoint = origin + (closestT*viewRay);
      surfaceNorm = (intersectPoint - closestSphere.center )/closestSphere.radius;
      if(closestSphere.textureIndex > 0)
      {
        texture tex = state->textures[closestSphere.textureIndex];
        u = atan2(surfaceNorm.x , surfaceNorm.z);
        v = acos(surfaceNorm.y)/PI;
        if(u < 0)
        {
          u = u + 2*PI;
        }
        u = u / (2*PI); 
        int i = round(u*(tex.width -1));
        int j = round(v*(tex.height -1));
        i = j*(tex.width*3) + i*3;

        mat.r = (float)tex.data[i]/255.0;
        mat.g = (float)tex.data[i+1]/255.0;
        mat.b = (float)tex.data[i+2]/255.0;
      }
      resultColor = phongLighting(viewRay, intersectPoint, surfaceNorm, mat, state, traceNumber, refractionIndexes);
      break;
  } //switch statement
  return resultColor;
}

int buildScene(renderState state, color *pixelBuffer)
{
  //error check static state values, handle error reporting inside function
  if(!checkStateVariables(&state))
  {
   return 0; 
  }
 
  // u =  viewdir x updir / magnitude
  vec3f u = crossProduct(state.viewdir, state.updir);
  if( u.x == 0 && u.y == 0 && u.z == 0)
  {
    std::cerr << "Error: please supply an updir and viewdir that are not parallel." << std::endl; 
    return 0;
  }

  u = normalizeVec3(u);
  // v = u x viewdir / magnitude
  vec3f v = crossProduct(u, state.viewdir);
  if (v.x == 0 && v.y == 0 && v.z == 0)
  {
    std::cerr << "This should not happen: u and viewdir calculated as parallel" << std::endl; 
    return 0;
  }

  //this should already be normal, but if only to remember that it should be: 
  v = normalizeVec3(v);
  //distance from eye to viewing window (is arbitrary)
  float d = 1;
  
  //compute viewing window size
  float viewWinH = degreesToRadians(state.fovv/2);
  viewWinH = 2*d*(tan( viewWinH));
  float aspectRatio = (float)state.imsizeW / state.imsizeH;
  float viewWinW = viewWinH * aspectRatio;

  // where n is a unit vector in the viewing direction: 
  vec3f n = normalizeVec3(state.viewdir); 

  // define image plane
  // p = center of image plane
  if(state.parallelMode)
  {
    d = 0;
  }
  
  point3f p = (state.eye + (d*n)); 
  point3f ul = p + (viewWinH/2)*v -  (viewWinW/2)*u;
  point3f ur = p + (viewWinH/2)*v +  (viewWinW/2)*u;
  point3f ll = p - (viewWinH/2)*v -  (viewWinW/2)*u;
  point3f lr = p - (viewWinH/2)*v +  (viewWinW/2)*u;

  // find vectors for traversing image->pixel space 
  //deltaXPlaneToPix  = ( upper right  - upper left of image plane ) / (output image width in pixels - 1)  
  state.deltaXPlaneToPix =  (ur - ul) / ((float) state.imsizeW - 1.0);
  state.deltaYPlaneToPix =  (ll - ul) / ((float) state.imsizeH - 1.0);

  //begin intersection tests
  vec3f viewRay;
  int currentPixel = 0;
  point3f eye = state.eye;
  //where i is the pixel row and j is the column:
  for(int i = 0; i < state.imsizeH; i++)
  {
    for(int j = 0; j < state.imsizeW; j++)
    { 
      // Upper left + pixel row * deltaY  +  pixel column * deltaX  = the point on the image plane corresponding to the current pixel
      point3f viewWindowPos = ul + i * state.deltaYPlaneToPix + j * state.deltaXPlaneToPix;
      viewRay = viewWindowPos - eye;
      viewRay = normalizeVec3(viewRay);
      if(state.parallelMode) 
      {
        eye = viewWindowPos;
        viewRay = n;
      }

	    float refractionIndexes[MAX_PHONG_RECURSIONS + 1] = {0};
      refractionIndexes[0] = 1.0f;
      //pass the state by reference, position/direction from the camera, 0 for the trace depth, and 
      //refraction index as an array (i.e., by reference)
      color resultColor = closestShapeAndLighting(&state, eye, viewRay, 0, refractionIndexes);
      pixelBuffer[currentPixel].r = resultColor.r;
      pixelBuffer[currentPixel].g = resultColor.g;
      pixelBuffer[currentPixel].b = resultColor.b;
      currentPixel++;
    }
  }
  //return okay when every pixel has been written into buffer:
  return 1;
}

int outputPPM(int width, int height, color* pixelBuffer, FILE* outputFile)
{
  //buffer to store the size line
  //note: this is large enough to store 8192 twice in string format, plus a space \n and \0
  char sizeLine[13];
  if(sprintf(sizeLine, "%d %d\n", width, height)<0)
  {
    std::cerr << "Error: could not create output size string" << std::endl; 
    return 0;
  }

  //create ppm header:
  if(fputs("P3\n", outputFile) == EOF ||
    fputs("#\n", outputFile) == EOF ||
    fputs(sizeLine, outputFile) == EOF ||
    fputs("255\n", outputFile) == EOF)
  {
    std::cerr << "Error: could not create whole output file header" << std::endl; 
    return 0;
  }

  int currentPixel = 0;
  color currentPixelColor;
  char pixelStr[13];
  int i = 0;
  for(; i < height*width ; i++)
  {
    currentPixelColor = pixelBuffer[currentPixel];
    int pixelR = (int) (currentPixelColor.r * 255);
    int pixelG = (int) (currentPixelColor.g * 255);
    int pixelB = (int) (currentPixelColor.b * 255);
	if (pixelR > 255 || pixelR < 0 || pixelG > 255 || pixelG < 0 || pixelB > 255 || pixelB < 0)
	{
		int test = 1;
	}
	if(sprintf(pixelStr, "%d %d %d\n", pixelR, pixelG, pixelB) < 0)
    {
      std::cerr << "Error: could not create output pixel color string" << std::endl; 
      return 0;
    }
    fputs(pixelStr, outputFile); 
    currentPixel++;   
  }
  return 1;
}